-- connctrl package of helper functions for the connctrl driver
--
-- API Functions:
--  is_dma_request_pending(devname)
--  is_dma_request_complete(devname)
--  is_dma_request_active(devname)
--  is_dma_blocked(devname)
--  is_dma_error(devname)
--  is_controller_halted(devname)
--  is_controller_in_reset(devname)
--  is_controller_enabled(devname)
--  is_host_initiated_dma_active(devname)
--  controller_reset(devname)
--  controller_run(devname)
--  initiate_host_dma_transaction(devname)
--  release_controller_reset(devname)
--  controller_stop(devname)
--  clear_host_dma_transaction(devname)
--  get_controller_memory_size(devname)
--  get_controller_map(devname)
--  get_controller_map_virtual_address(devname)
--  get_dma_mapping(devname, size)
--  release_dma_mapping(dma_map)
--  get_dma_mapping_physical_address(dma_map)
--  get_dma_mapping_virtual_address(dma_map)
--  close_device(devname)
--  dma_host_transaction(devname, offset_in_bar, dma_map, offset_in_dma_page,
--                       size, is_fpga_to_host)
--  load_microcode(devname, microcode_file)
--  test_for_interrupt(devname)
--  wait_for_interrupt(devname)
--  is_spi_control_register_present()
--  spi_transaction(byte)
--  end_spi_command()
--

local c = require 'system_config'
if c.device_type == 'edaqxr' then return require 'edaqctrl' end

local u = require 'unix'
local bit = require 'bit'
local ffi = require 'ffi'
local lja = require 'ljarmada'

local connctrl = {}

-- Store file descriptor indexed by device name
local open_devices = {}

-- Store map, map size, and command reg shadow contents indexed by file desc
local open_bar_mappings = {}

-- Store physical address, size, fd, and cache line size indexed by map
local open_dma_mappings = {}

function connctrl.dump_internal_tables()
    print("Open Devices")
    print("File", "Device")
    for k,v in pairs(open_devices) do print(v, k) end
    print("Open BAR Mappings")
    print("File", "Map", "", "", "Size", "Command Register Shadow")
    for k,v in pairs(open_bar_mappings) do
        print(k, v.map, v.sz, string.format("%08X", v.shadow))
    end
    print("Open DMA Mappings")
    print("File", "Map", "", "", "Phys Addr", "Size", "Cache Line Size")
    for k,v in pairs(open_dma_mappings) do
        print(v.fd, k, string.format("0x%08X", v.pa), v.sz, v.cachelinesize)
    end
end

function connctrl.dump_exception_info_s(devname)
    local info = ""
    local function print(s) info = info .. s .. '\r\n' end
    local function iowrite(s) info = info .. s end
    local bar_map = connctrl.get_controller_map(devname)
    local exception_name = { [0] = "Program End",
                             [2] = "Bus Error",
                             [6] = "Alignment",
                             [7] = "Illegal Instruction",
                             [8] = "Unhandled Interrupt",
                             [11] = "Range Error",
                             [12] = "Unhandled Syscall",
                             [13] = "Break",
                             [14] = "Trap" }
    local fp = bar_map:r32(8)
    local memsize = bar_map:r32(0)
    print("\r\nException Dump")
    local exception_number = bar_map:r32(4)
    if exception_name[exception_number] then
        print(string.format("Type            %s", exception_name[exception_number]))
    else
        print(string.format("Type            %d", exception_number))
    end
    print(string.format("Memory Size     %d", memsize))
    print(string.format("Frame Pointer   %08X", fp))
    print(string.format("Fault Address   %08X", bar_map:r32(12)))
    print(string.format("DMA Remote Addr %08X", bar_map:r32(0xFFFF0)))
    print(string.format("DMA Local Addr  %08X", bar_map:r32(0xFFFF8)))
    print(string.format("DMA Parameters  %08X", bar_map:r32(0xFFFFC)))
    fp = bit.band(fp, memsize - 1)
    local sp
    for i=0,31,4 do
        local rval
        local rn
        for j=0,3 do
            rn = i + j
            if rn == 0 then rval = 0
            else if rn == 1 then rval = bar_map:r32(8) + 128 sp = rval
            else rval = bar_map:r32(fp + ((rn - 2) * 4)) end end
            iowrite(string.format("r%-2d  %08X   ", rn, rval))
        end
        print("")
    end
    print("Stack Dump")
    for i=0,256,16 do
        if sp + i >= 0x100000 then break end
        iowrite(string.format("%08X:  ", sp + i))
        for j=0,12,4 do
            if sp + i + j >= 0x100000 then break end
            iowrite(string.format("%08X  ", bar_map:r32(sp + i + j)))
        end
        print("")
    end
    if bar_map:r16(0x1c) ~= 0 then
        print(connctrl.dump_trace_buffer_s(devname))
    end
    return info
end

function connctrl.dump_exception_info(devname)
    print(connctrl.dump_exception_info_s(devname))
end

function connctrl.dump_trace_buffer_s(devname)
    local info = ""
    local function print(s) info = info .. s .. '\r\n' end
    local function iowrite(s) info = info .. s end
    local bar_map = connctrl.get_controller_map(devname)
    local trace_buffer_size = bar_map:r16(0x1c)
    if trace_buffer_size == 0 then
        print("No Trace Information")
        return info
    end
    local trace_next_entry = bar_map:r16(0x1e)
    local trace_buffer = bar_map:r32(0x18)
    print("Trace Buffer Dump")
    local entries_processed = 0
    local extra_words = 0
    while entries_processed < trace_buffer_size do
        trace_next_entry = trace_next_entry - 1
        if trace_next_entry < 0 then
            trace_next_entry = trace_buffer_size - 1
        end
        local entry = bar_map:r32(trace_buffer + (trace_next_entry * 4))
        if extra_words == 0 then
            extra_words = bit.band(bit.rshift(entry, 24), 3)
            local tag = bit.rshift(entry, 26)
            local data1 = bit.band(entry, 0x00FFFFFF)
            if tag == 0 then return info end
            iowrite(string.format("%02X  %06X", tag, data1))
        else
            iowrite(string.format("  %08X", entry))
            extra_words = extra_words - 1
        end
        if extra_words == 0 then print("") end
        entries_processed = entries_processed + 1
    end
    return info
end

function connctrl.dump_trace_buffer(devname)
    print(connctrl.dump_trace_buffer_s(devname))
end

function connctrl.check_for_exception(devname)
    if connctrl.is_controller_halted(devname) then
        local info = string.format("Controller %s has halted.\r", devname)
        info = info .. connctrl.dump_exception_info_s(devname)
        connctrl.controller_reset(devname)
        error(info)
    end
end

local function get_fd(devname)
    local fd = open_devices[devname]
    if not fd then
        fd = assert(u.open(devname, u.O_RDWR))
        open_devices[devname] = fd
    end
    return fd
end

local function get_bar_map_entry(devname)
    local fd = get_fd(devname)
    local map_entry = open_bar_mappings[fd]
    if not map_entry then
        local map, va = assert(u.mmap(0x101000, fd, 0))
        map_entry = { map = map, va = va }
        open_bar_mappings[fd] = map_entry
    end
    return map_entry
end

local function get_status_register(devname)
    local map_entry = get_bar_map_entry(devname)
    return map_entry.map:r32(0x100000)
end

local function get_cmd_register(devname)
    local map_entry = get_bar_map_entry(devname)
    if map_entry.shadow then return map_entry.shadow end
    map_entry.shadow = map_entry.map:r32(0x100004)
    return map_entry.shadow
end

local function set_cmd_register(devname, value)
    local map_entry = get_bar_map_entry(devname)
    map_entry.shadow = value
    map_entry.map:w32(0x100004, value)
end

function connctrl.is_dma_request_pending(devname)
    return bit.band(get_status_register(devname), 2) ~= 0
end

function connctrl.is_dma_request_complete(devname)
    return bit.band(get_status_register(devname), 4) ~= 0
end

function connctrl.is_dma_request_active(devname)
    return bit.band(get_status_register(devname), 8) ~= 0
end

function connctrl.is_dma_blocked(devname)
    return bit.band(get_status_register(devname), 16) ~= 0
end

function connctrl.is_dma_error(devname)
    return bit.band(get_status_register(devname), 32) ~= 0
end

function connctrl.is_controller_halted(devname)
    return bit.band(get_status_register(devname), 64) ~= 0
end

function connctrl.is_controller_in_reset(devname)
    return bit.band(get_cmd_register(devname), 1) ~= 0
end

function connctrl.is_controller_enabled(devname)
    return bit.band(get_cmd_register(devname), 2) ~= 0
end

function connctrl.is_host_initiated_dma_active(devname)
    return bit.band(get_cmd_register(devname), 4) ~= 0
end

function connctrl.controller_reset(devname)
    set_cmd_register(devname, 1)
end

function connctrl.controller_run(devname)
    set_cmd_register(devname, bit.bor(get_cmd_register(devname), 2))
end

function connctrl.initiate_host_dma_transaction(devname)
    set_cmd_register(devname, bit.bor(get_cmd_register(devname), 4))
end

function connctrl.release_controller_reset(devname)
    set_cmd_register(devname, bit.band(get_cmd_register(devname), bit.bnot(1)))
end

function connctrl.controller_stop(devname)
    set_cmd_register(devname, bit.band(get_cmd_register(devname), bit.bnot(2)))
end

function connctrl.clear_host_dma_transaction(devname)
    set_cmd_register(devname, bit.band(get_cmd_register(devname), bit.bnot(4)))
end

function connctrl.get_controller_memory_size(devname)
    local map_entry = get_bar_map_entry(devname)
    local j = 1
    local n = 1
    local loc0_contents, test_contents, chkval
    loc0_contents = map_entry.map:r8(0)
    map_entry.map:w8(0, 0)
    while j < 0x100000 do
        test_contents = map_entry.map:r8(j)
        map_entry.map:w8(j, n)
        chkval = map_entry.map:r8(0)
        map_entry.map:w8(j, test_contents)
        n = n + 1
        if chkval ~= 0 then break end
        j = j * 2
    end
    map_entry.map:w8(0, loc0_contents)
    return j
end

function connctrl.get_controller_map(devname)
    local map_entry = get_bar_map_entry(devname)
    return map_entry.map
end

function connctrl.get_controller_map_virtual_address(devname)
    local map_entry = get_bar_map_entry(devname)
    return map_entry.va
end

function connctrl.get_dma_mapping(devname, size)
    local fd = get_fd(devname)
    if size == 0x101000 then size = size + 4 end     -- Avoid 'magic' size
    local map, va = assert(u.mmap(size, fd, 0))
    lja.map_flush(ffi.cast('char *', va), 8, 8)
    open_dma_mappings[map] = { pa=map:r32(0), sz=size, fd=fd, va=va,
                               cachelinesize=map:r32(4) }
    map:w32(0, 0)   -- Hide physical address of mapping
    map:w32(4, 0)   -- Hide cache line size of mapping
    local mi = open_dma_mappings[map]
    lja.map_flush(ffi.cast('char *', mi.va), mi.cachelinesize, mi.sz)
    return map
end

function connctrl.flush_dma_region_by_offset(map, length, offset)
    local map_info = assert(open_dma_mappings[map])
    lja.map_flush(ffi.cast('char *', map_info.va + tonumber(offset)),
        map_info.cachelinesize, tonumber(length))
end

function connctrl.flush_dma_region_by_ptr(map, ptr, length)
    local map_info = assert(open_dma_mappings[map])
    lja.map_flush(ffi.cast('char *', ptr), map_info.cachelinesize,
        tonumber(length))
end

function connctrl.flush_entire_dma_region(map)
    local map_info = assert(open_dma_mappings[map])
    lja.map_flush(ffi.cast('char *', map_info.va), map_info.cachelinesize,
        map_info.sz)
end

function connctrl.release_dma_mapping(dma_map)
    open_dma_mappings[dma_map] = nil
end

function connctrl.get_dma_mapping_physical_address(dma_map)
    return open_dma_mappings[dma_map].pa
end

function connctrl.get_dma_mapping_virtual_address(dma_map)
    return open_dma_mappings[dma_map].va
end

function connctrl.close_device(devname)
    local fd = open_devices[devname]
    if fd then
        open_devices[devname] = nil
        open_bar_mappings[fd] = nil
        u.close(fd)
    end
end

local function compute_first_be(address_2lsbs, length)
    local first_be
    if length == 0 then first_be = 0x0
    elseif address_2lsbs == 0 and length == 1 then first_be = 0x1
    elseif address_2lsbs == 1 and length == 1 then first_be = 0x2
    elseif address_2lsbs == 2 and length == 1 then first_be = 0x4
    elseif address_2lsbs == 3 and length >= 1 then first_be = 0x8
    elseif address_2lsbs == 0 and length == 2 then first_be = 0x3
    elseif address_2lsbs == 1 and length == 2 then first_be = 0x6
    elseif address_2lsbs == 2 and length >= 2 then first_be = 0xC
    elseif address_2lsbs == 0 and length == 3 then first_be = 0x7
    elseif address_2lsbs == 1 and length >= 3 then first_be = 0xE
    else first_be = 0xF end
    return first_be
end

local function compute_last_be_and_dw(address_2lsbs, length)
    local last_be
    local dw = 1
    if (address_2lsbs == 0 and length <= 4) or
       (address_2lsbs == 1 and length <= 3) or
       (address_2lsbs == 2 and length <= 2) or
       (address_2lsbs == 3 and length <= 1) then last_be = 0x0
    else
        length = length - (4 - address_2lsbs)
        if length % 4 == 0 then last_be = 0xF
        elseif length % 4 == 1 then last_be = 0x1
        elseif length % 4 == 2 then last_be = 0x3
        else last_be = 0x7 end
        dw = dw + math.floor((length + 3) / 4)
    end
    if dw == 1024 then dw = 0 end
    return last_be, dw
end

local function compute_dma_control_word(address_2lsbs, length, is_fpga_to_host)
    local control_word
    local first_be = compute_first_be(address_2lsbs, length)
    local last_be, dw = compute_last_be_and_dw(address_2lsbs, length)
    control_word = last_be * 0x100000 + first_be * 0x10000 + dw
    if is_fpga_to_host then control_word = control_word + 0x400 end
    return control_word
end

function connctrl.dma_host_transaction(devname, offset_in_bar,
                                       dma_map, offset_in_dma_page,
                                       length, is_fpga_to_host)
    if offset_in_bar % 4 ~= offset_in_dma_page % 4 then
        return("Unaligned DMA Requested")
    end

    local bar_map = connctrl.get_controller_map(devname)

    -- write DMA parameters
    bar_map:w32(0xFFFF0, open_dma_mappings[dma_map].pa + offset_in_dma_page)
    bar_map:w32(0xFFFF8, offset_in_bar)
    bar_map:w32(0xFFFFC, compute_dma_control_word(offset_in_bar % 4, length,
                                                  is_fpga_to_host))

    -- write initiate dma from host command
    connctrl.initiate_host_dma_transaction(devname)
    bar_map:r32(0xFFFFC) -- Benign readback to cause short delay
    local result = connctrl.is_dma_request_complete(devname)
    connctrl.clear_host_dma_transaction(devname)
    return result
end

function connctrl.load_microcode(devname, microcode_file)
    local bar_map = connctrl.get_controller_map(devname)
    local bar_mem_size = connctrl.get_controller_memory_size(devname)

    local fd = assert(u.open(microcode_file, u.O_RDONLY))
    local microcode_bytes = assert(u.read(fd, bar_mem_size))

    if #microcode_bytes > bar_mem_size then
        error("Microcode file size exceeds BAR1 memory size")
    end

    bar_map:ws(0, microcode_bytes)

    u.close(fd)
end

function connctrl.test_for_interrupt(devname)
    local fd = assert(u.open(devname, u.O_RDWR + u.O_NONBLOCK))

    local buf, es, ec = u.read(fd, 1)

    u.close(fd)

    if ec == 11 then return false end

    if buf then
        return true
    end

    error(es)
end

function connctrl.wait_for_interrupt(devname)
    local fd = get_fd(devname)

    u.read(fd, 1)
end

function connctrl.clear_interrupts(devname)
    local fd = assert(u.open(devname, u.O_RDWR + u.O_NONBLOCK))

    while true do
        local buf, es, ec = u.read(fd, 1)

        if ec == 11 then
            u.close(fd)
            return
        end

        if not buf then
            error(es)
        end
    end
end

function connctrl.is_spi_control_register_present()
    local map_entry = get_bar_map_entry('/dev/cca0')
    map_entry.map:w32(0x10000C, 0)
    local spi_cs_should_be_zero = map_entry.map:r32(0x10000C)
    map_entry.map:w32(0x10000C, 0x80000000)
    local spi_cs_should_be_one = map_entry.map:r32(0x10000C)
    map_entry.map:w32(0x10000C, 0)
    return((bit.band(spi_cs_should_be_zero, 0x80000000) == 0) and
            (bit.band(spi_cs_should_be_one, 0x80000000) ~= 0))
end

function connctrl.spi_transaction(byte)
    local map_entry = get_bar_map_entry('/dev/cca0')
    map_entry.map:w32(0x10000C, bit.bor(bit.band(byte, 0xFF), 0x80000000))
    local dummy = map_entry.map:r32(0x10000C)
    local result = map_entry.map:r32(0x10000C)
    assert(bit.band(result, 0x01000000) == 0)
    return bit.band(result, 0xFF)
end

function connctrl.end_spi_command()
    local map_entry = get_bar_map_entry('/dev/cca0')
    map_entry.map:w32(0x10000C, 0)
end

return connctrl

