-- dio package
--
-- Provides an API to a DIO channel
--
-- API Functions:
--  dio:new(devname, firmware_path, master_freq)
--  dio:poll()
--  dio:ts_offset(ts)
--  dio:setup_sync_ts()
--  dio:get_sync_ts()
--  dio:get_current_ts() [DEBUG]
--  dio:txdr_power_on_5v()
--  dio:txdr_power_on_12v()
--  dio:txdr_power_off()
--  dio:hc_start(sample_rate_numerator, sample_rate_denominator, start_time,
--               samples_per_bd, num_bds, hysteresis_args)
--  dio:hc_stop()
--  dio:hc_set_thresholds(upper_voltage, lower_voltage)
--  dio:ic_start(sample_rate_numerator, sample_rate_denominator, start_time,
--               samples_per_bd, num_bds, ic_args)
--  dio:ic_stop()
--  dio:pc_start(sample_rate_numerator, sample_rate_denominator, start_time,
--               samples_per_bd, num_bds, pc_args)
--  dio:pc_stop()
--  dio:qc_start(sample_rate_numerator, sample_rate_denominator, start_time,
--               samples_per_bd, num_bds, quadrature_args)
--  dio:qc_stop()
--  dio:os_start(samples_per_bd, num_bds)
--  dio:os_append(ts_array, initial_state, output_bit)
--  dio:os_stop(output_bit)
--  dio:event_setup(event_args)
--  dio:event_start(start_time, heartbeat_interval,
--                  samples_per_bd, num_bds)
--  dio:event_stop()
--  dio:set_output(num, on)
--  dio:get_cable_detect_state()
--
--
-- Note on all commands with output_bit as a parameter:
--   output_bit is either 0 or 1 representing an output signal pin.
--
-- Notes on event_setup(event_args) parameters:
--   Set the following parameter names to true or non-zero to enable.
--       cable_unplugged
--       cable_plugged
--       index_mark_fall
--       sync_input0_fall
--       sync_input1_fall
--       sync_input2_fall
--       index_mark_rise
--       sync_input0_rise
--       sync_input1_rise
--       sync_input2_rise
--       output2_off
--       output2_on
--       output2_forced_off
--       output2_forced_on
--       output1_off
--       output1_on
--       output1_forced_off
--       output1_forced_on
--
--   For example: event_setup({cable_unplugged = true})
--   All parameters not listed are disabled by default.
--
-- Notes on hc_start(..., hysteresis_args) parameters:
--   Set the following parameter names to appropriate values:
--           dynamic_mode          true to enable dynamic mode
--           decay_rate            (Amount in volts that a peak decays
--                                  at the decay algorithm run rate)
--           decay_divider         (Divider from programmed sample rate,
--                                  sets the decay algorithm run rate)
--           dac_update_divider    (Divider from programmed sample rate,
--                                  determines rate of DAC updates)
--           threshold_ratio_lo2hi (Ratio between high and low peaks to
--                                  set the hi threshold, 0.0 to 1.0)
--           threshold_ratio_hi2lo (Ratio between high and low peaks to
--                                  set the lo threshold, 0.0 to 1.0)
--           initial_threshold_hi  (Initial dynamic threshold hi voltage)
--           initial_threshold_lo  (Initial dynamic threshold lo voltage)
--           threshold_hi          (Static threshold hi voltage)
--           threshold_lo          (Static threshold lo voltage)
--
--   For example: hc_start(..., {threshold_hi = 2.4, threshold_lo = 1.4})
--   If dynamic_mode is missing or set to false, the threshold_lo and
--   threshold_hi parameters are required.  If dynamic_mode is set to true,
--   all the other parameters are required.  Divider parameters must be integers
--   greater than zero but will be rounded down to the nearest integer and
--   ratios less than one will be clipped to one.  Ratio parameters must be
--   between 0.0 and 1.0 but will be clipped.  Voltage parameters must be
--   between -5.0 and 5.0 but will be clipped.
--   Missing or malformed parameters will abort the function with a Lua error.
--
-- Notes on qc_start(..., quadrature_args) parameters:
--   Set the following parameter names to true or non-zero to enable.
--           swap_a_b
--           invert_index
--           invert_b
--           invert_a
--           count_down_b_fall_a_one
--           count_down_b_rise_a_zero
--           count_down_a_fall_b_zero
--           count_down_a_rise_b_one
--           count_up_b_fall_a_zero
--           count_up_b_rise_a_one
--           count_up_a_fall_b_one
--           count_up_a_rise_b_zero
--
--   For example: qc_start(..., {invert_b = true})
--   All parameters not listed are disabled by default.
--   Note that some of the count* parameters MUST be enabled otherwise
--   the counter will not move.
--
-- Notes on oc_append():
--   Provide an array of timestamps where transitions take place along with
--   an initial value (0 = open circuit, 1 = closed circuit/shorted).  The
--   function will attempt to queue up as many transitions as possible and
--   return the number of transitions queued.
--


local __ = require '__'
local log = require 'log'
local cc = require 'connctrl'
local ffi = require 'ffi'
local rb = require 'ring_buffers'
local ch = require 'channel'

local dio = setmetatable({}, { __index = ch })
local f = assert(io.open("/usr/include/fpga-firmware/dio_structures.h", "r"))
local h = f:read("*all")
f:close()
ffi.cdef(h)


function dio:new(devname, firmware_path, master_freq)
    local dio_ch = self:new_channel(devname, firmware_path, "dio_cmd_s")
    dio_ch.master_freq = master_freq
    dio_ch.os_rb_size = {}
    return dio_ch
end

function dio:ts_offset(ts)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_TS_OFFSET
    diocmd.prm.ts_offset.ts_adjust = ts
    return self:send_command_buffer(diocmd)
end

function dio:setup_sync_ts()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_SETUP_SYNC_TS
    return self:send_command_buffer(diocmd)
end

function dio:get_sync_ts()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_GET_SYNC_TS
    return self:send_command_buffer(diocmd)
end

function dio:get_current_ts()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_GET_TS
    return self:send_command_buffer(diocmd)
end

function dio:txdr_power_on_5v()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_TXDR_POWER_ON_5V
    return self:send_command_buffer(diocmd)
end

function dio:txdr_power_on_12v()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_TXDR_POWER_ON_12V
    return self:send_command_buffer(diocmd)
end

function dio:txdr_power_off()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_TXDR_POWER_OFF
    return self:send_command_buffer(diocmd)
end

local function compute_bit_value(voltage)
    voltage = tonumber(voltage)
	if voltage > 5 then voltage = 5 end
	if voltage < -5 then voltage = -5 end
	local dac_voltage = voltage * 0.2 + 1.25
    local value
    value = math.floor(dac_voltage / 2.5 * 65535 + 0.5)
	return value
end

function dio:hc_start(sample_rate_numerator, sample_rate_denominator,
                      start_time, samples_per_bd, num_bds,
                      hysteresis_args)
    local decay_rate, decay_divider, dac_update_divider
    local threshold_ratio_lo2hi, threshold_ratio_hi2lo
    local threshold_hi, threshold_lo, dynamic_mode

    -- Check all arguments first
    if hysteresis_args.dynamic_mode then
        dynamic_mode = 1
        assert(hysteresis_args.decay_rate)
        assert(hysteresis_args.decay_divider)
        assert(hysteresis_args.dac_update_divider)
        assert(hysteresis_args.threshold_ratio_lo2hi)
        assert(hysteresis_args.threshold_ratio_hi2lo)
        assert(hysteresis_args.initial_threshold_hi)
        assert(hysteresis_args.initial_threshold_lo)
        threshold_hi = compute_bit_value(hysteresis_args.initial_threshold_hi)
        threshold_lo = compute_bit_value(hysteresis_args.initial_threshold_lo)
        if hysteresis_args.decay_rate >= 10.0 then
            decay_rate = 9.99999
        else if hysteresis_args.decay_rate < 0.0 then
            decay_rate = 0.0
        else
            decay_rate = hysteresis_args.decay_rate
        end end
        decay_rate = math.floor(decay_rate / 10 * 65536)
        decay_divider = math.floor(hysteresis_args.decay_divider)
        if decay_divider < 1 then decay_divider = 1 end
        if decay_divider > 4096 then decay_divider = 4096 end
        dac_update_divider = math.floor(hysteresis_args.dac_update_divider)
        if dac_update_divider < 1 then dac_update_divider = 1 end
        if dac_update_divider > 4096 then dac_update_divider = 4096 end
        if hysteresis_args.threshold_ratio_lo2hi > 1.0 then
            threshold_ratio_lo2hi = 1.0
        else if hysteresis_args.threshold_ratio_lo2hi < 0.0 then
            threshold_ratio_lo2hi = 0.0
        else
            threshold_ratio_lo2hi = hysteresis_args.threshold_ratio_lo2hi
        end end
        threshold_ratio_lo2hi = math.floor(threshold_ratio_lo2hi * 16384)
        if hysteresis_args.threshold_ratio_hi2lo > 1.0 then
            threshold_ratio_hi2lo = 1.0
        else if hysteresis_args.threshold_ratio_hi2lo < 0.0 then
            threshold_ratio_hi2lo = 0.0
        else
            threshold_ratio_hi2lo = hysteresis_args.threshold_ratio_hi2lo
        end end
        threshold_ratio_hi2lo = math.floor(threshold_ratio_hi2lo * 16384)
    else
        dynamic_mode = 0
        assert(hysteresis_args.threshold_hi)
        assert(hysteresis_args.threshold_lo)
        threshold_hi = compute_bit_value(hysteresis_args.threshold_hi)
        threshold_lo = compute_bit_value(hysteresis_args.threshold_lo)
        decay_rate = 0
        decay_divider = 1
        dac_update_divider = 1
        threshold_ratio_lo2hi = 0
        threshold_ratio_hi2lo = 0
    end

    -- Now fill out a command structure
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_HC_START
    diocmd.prm.hc_start.dynamic_mode = dynamic_mode
    diocmd.prm.hc_start.start_time = start_time
    self:calculate_scg_parameters(diocmd.prm.hc_start.sample_rate,
                                  sample_rate_numerator,
                                  sample_rate_denominator,
                                  self.master_freq)
    local pa = self:add_ring_buffer('HC',
                                    "hc_buffer_t["..samples_per_bd.."]",
                                    "hc_buffer_t", num_bds)
    diocmd.prm.hc_start.phys_address = pa
    diocmd.prm.hc_start.threshold_hi = threshold_hi
    diocmd.prm.hc_start.threshold_lo = threshold_lo
    diocmd.prm.hc_start.decay_rate = decay_rate
    diocmd.prm.hc_start.decay_divider = decay_divider - 1
    diocmd.prm.hc_start.dac_update_divider = dac_update_divider - 1
    diocmd.prm.hc_start.threshold_ratio_lo2hi = threshold_ratio_lo2hi
    diocmd.prm.hc_start.threshold_ratio_hi2lo = threshold_ratio_hi2lo
    return self:send_command_buffer(diocmd)
end

function dio:hc_stop()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_HC_STOP
    return self:send_command_buffer(diocmd)
end

function dio:hc_set_thresholds(upper_voltage, lower_voltage)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_HC_SET_THRESHOLDS
    diocmd.prm.threshold_setup.threshold_hi = compute_bit_value(upper_voltage)
    diocmd.prm.threshold_setup.threshold_lo = compute_bit_value(lower_voltage)
    return self:send_command_buffer(diocmd)
end

function dio:ic_start(sample_rate_numerator, sample_rate_denominator,
                      start_time, samples_per_bd, num_bds, ic_args)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_IC_START
    diocmd.prm.ic_start.start_time = start_time
    self:calculate_scg_parameters(diocmd.prm.ic_start.sample_rate,
                                  sample_rate_numerator,
                                  sample_rate_denominator,
                                  self.master_freq)
    local pa = self:add_ring_buffer('IC',
                                    "ic_buffer_t["..samples_per_bd.."]",
                                    "ic_buffer_t", num_bds)
    diocmd.prm.ic_start.phys_address = pa
    return self:send_command_buffer(diocmd)
end

function dio:ic_stop()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_IC_STOP
    return self:send_command_buffer(diocmd)
end

function dio:pc_start(sample_rate_numerator, sample_rate_denominator,
                      start_time, samples_per_bd, num_bds, pc_args)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_PC_START
    diocmd.prm.pc_start.start_time = start_time
    self:calculate_scg_parameters(diocmd.prm.pc_start.sample_rate,
                                  sample_rate_numerator,
                                  sample_rate_denominator,
                                  self.master_freq)
    local pa = self:add_ring_buffer('PC',
                                    "pc_buffer_t["..samples_per_bd.."]",
                                    "pc_buffer_t", num_bds)
    diocmd.prm.pc_start.phys_address = pa
    return self:send_command_buffer(diocmd)
end

function dio:pc_stop()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_PC_STOP
    return self:send_command_buffer(diocmd)
end

function dio:qc_start(sample_rate_numerator, sample_rate_denominator,
                      start_time, samples_per_bd, num_bds, quadrature_args)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_QC_START
    diocmd.prm.qc_start.start_time = start_time
    self:calculate_scg_parameters(diocmd.prm.qc_start.sample_rate,
                                  sample_rate_numerator,
                                  sample_rate_denominator,
                                  self.master_freq)
    local pa = self:add_ring_buffer('QC',
                                    "qc_buffer_t["..samples_per_bd.."]",
                                    "qc_buffer_t", num_bds)
    diocmd.prm.qc_start.phys_address = pa
    if quadrature_args.swap_a_b then diocmd.prm.qc_start.swap_a_b = 1 end
    if quadrature_args.invert_index then diocmd.prm.qc_start.invert_index = 1 end
    if quadrature_args.invert_b then diocmd.prm.qc_start.invert_b = 1 end
    if quadrature_args.invert_a then diocmd.prm.qc_start.invert_a = 1 end
    if quadrature_args.count_down_b_fall_a_one then
        diocmd.prm.qc_start.count_down_b_fall_a_one = 1
    end
    if quadrature_args.count_down_b_rise_a_zero then
        diocmd.prm.qc_start.count_down_b_rise_a_zero = 1
    end
    if quadrature_args.count_down_a_fall_b_zero then
        diocmd.prm.qc_start.count_down_a_fall_b_zero = 1
    end
    if quadrature_args.count_down_a_rise_b_one then
        diocmd.prm.qc_start.count_down_a_rise_b_one = 1
    end
    if quadrature_args.count_up_b_fall_a_zero then
        diocmd.prm.qc_start.count_up_b_fall_a_zero = 1
    end
    if quadrature_args.count_up_b_rise_a_one then
        diocmd.prm.qc_start.count_up_b_rise_a_one = 1
    end
    if quadrature_args.count_up_a_fall_b_one then
        diocmd.prm.qc_start.count_up_a_fall_b_one = 1
    end
    if quadrature_args.count_up_a_rise_b_zero then
        diocmd.prm.qc_start.count_up_a_rise_b_zero = 1
    end
    return self:send_command_buffer(diocmd)
end

function dio:qc_stop()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_QC_STOP
    return self:send_command_buffer(diocmd)
end

function dio:os_start(samples_per_bd, num_bds, output_bit)
    local pa = self:add_ring_buffer('OS'..output_bit,
                                    "os_buffer_t["..samples_per_bd.."]",
                                    "os_buffer_t", num_bds)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_OS_START
    diocmd.prm.os_start.phys_address = pa
    diocmd.prm.os_start.output_bit = output_bit
    self.os_rb_size['OS'..output_bit] = samples_per_bd
    return self:send_command_buffer(diocmd)
end

function dio:os_append(ts_array, initial_state, output_bit)
    local rb_name = 'OS'..output_bit
    local samples_per_bd = self.os_rb_size[rb_name]
    local total_samples = #ts_array
    local index = 1
    local current_state = initial_state
    local samples_queued = 0
    while total_samples > 0 do
        local samples_in_buffer
        if total_samples > samples_per_bd then
            samples_in_buffer = samples_per_bd
        else
            samples_in_buffer = total_samples
        end
        local os_buffer = ffi.new("os_buffer_t", samples_in_buffer)
        for i=index,index + samples_in_buffer - 1 do
            os_buffer[i - index].ts = ts_array[i]
            os_buffer[i - index].match_closed = current_state
            if current_state == 0 then
                current_state = 1
            else
                current_state = 0
            end
        end
        local buffer_size = samples_in_buffer * ffi.sizeof("os_buffer_t")
        if self:send_buffer(os_buffer, rb_name, buffer_size) then
            total_samples = total_samples - samples_in_buffer
            samples_queued = samples_queued + samples_in_buffer
        else
            return samples_queued
        end
    end
    return samples_queued
end

function dio:os_stop(output_bit)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_OS_STOP
    diocmd.prm.stop_channel = output_bit
    return self:send_command_buffer(diocmd)
end

function dio:event_setup(arg)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_EVENT_SETUP
    if arg.cable_unplugged then diocmd.prm.events.cable_unplugged = 1 end
    if arg.cable_plugged then diocmd.prm.events.cable_plugged = 1 end
    if arg.index_mark_fall then diocmd.prm.events.index_mark_fall = 1 end
    if arg.sync_input0_fall then diocmd.prm.events.sync_input0_fall = 1 end
    if arg.sync_input1_fall then diocmd.prm.events.sync_input1_fall = 1 end
    if arg.sync_input2_fall then diocmd.prm.events.sync_input2_fall = 1 end
    if arg.index_mark_rise then diocmd.prm.events.index_mark_rise = 1 end
    if arg.sync_input0_rise then diocmd.prm.events.sync_input0_rise = 1 end
    if arg.sync_input1_rise then diocmd.prm.events.sync_input1_rise = 1 end
    if arg.sync_input2_rise then diocmd.prm.events.sync_input2_rise = 1 end
    if arg.output1_off then diocmd.prm.events.match_closed0 = 1 end
    if arg.output1_on then diocmd.prm.events.match_open0 = 1 end
    if arg.output1_forced_off then diocmd.prm.events.force_closed0 = 1 end
    if arg.output1_forced_on then diocmd.prm.events.force_open0 = 1 end
    if arg.output2_off then diocmd.prm.events.match_closed1 = 1 end
    if arg.output2_on then diocmd.prm.events.match_open1 = 1 end
    if arg.output2_forced_off then diocmd.prm.events.force_closed1 = 1 end
    if arg.output2_forced_on then diocmd.prm.events.force_open1 = 1 end
    return self:send_command_buffer(diocmd)
end

function dio:event_start(start_time, heartbeat_interval,
                         samples_per_bd, num_bds)
    local pa = self:add_ring_buffer('EV',
                                    "dio_ev_buffer_t["..samples_per_bd.."]",
                                    "dio_ev_buffer_t", num_bds)
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_EVENT_START
    diocmd.prm.ev_start.phys_address = pa
    diocmd.prm.ev_start.start_time = start_time
    self:calculate_scg_parameters(diocmd.prm.ev_start.heartbeat_interval,
                                  heartbeat_interval, 1, self.master_freq)
    return self:send_command_buffer(diocmd)
end

function dio:event_stop()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_EVENT_STOP
    return self:send_command_buffer(diocmd)
end

function dio:set_output(num, on)
    local diocmd = ffi.new("dio_cmd_s")
    if num == 1 then
        if on then
            diocmd.cmd = ffi.C.DIO_SET_OUTPUT0_OPEN
        else
            diocmd.cmd = ffi.C.DIO_SET_OUTPUT0_CLOSED
        end
    elseif num == 2 then
        if on then
            diocmd.cmd = ffi.C.DIO_SET_OUTPUT1_OPEN
        else
            diocmd.cmd = ffi.C.DIO_SET_OUTPUT1_CLOSED
        end
    else
        log:warn(__"Digital Output [{0}] NOT VALID", num)
        return
    end
    return self:send_command_buffer(diocmd)
end

function dio:get_cable_detect_state()
    local diocmd = ffi.new("dio_cmd_s")
    diocmd.cmd = ffi.C.DIO_GET_CABLE_SENSE
    return self:send_command_buffer(diocmd)
end

return dio
