
local u = require 'unix'
local b = require 'bit'

local fpga = {}

local function reprogram_fpga(svf_file)
    if svf_file then
        local f = assert(io.popen("/sbin/lsmod | grep connctrl", 'r'))
        local load_status = f:read()
        f:close()
        if load_status then
            return "Cannot proceed, connctrl driver is loaded"
        end
    end

    local fd = assert(u.open('/dev/mem', u.O_RDWR))
    local map = assert(u.mmap(0x100000, fd, 0xe0000000))
    assert(u.close(fd))

    local save_SICRL = map:r32(0x00114)
    local save_SICRH = map:r32(0x00118)

    -- Set USB_B group to GPIO (for GPIO2[11] => TMS)
    -- Set GPIO1_L to GPIO (for GPIO1[11] => TDO)
    -- Set GPIO1_J to GPIO (for GPIO1[10] => TDI)
    local new_SICRL = b.bor(b.band(save_SICRL, 0xcfff4fff), 0x00000000)

    -- Set GPIO2_F to GPIO (for GPIO2[17] => TCK)
    local new_SICRH = b.bor(b.band(save_SICRH, 0xfffffcff), 0x00000100)

    -- Reconfigure GPIOs
    map:w32(0x00114, new_SICRL)
    map:w32(0x00118, new_SICRH)

    local tmpfile = io.open('/tmp/jtag.cmd', 'w+')
    tmpfile:write("cable gpio tdi=203 tdo=234 tck=209 tms=235\n")
    tmpfile:write("detect\n")
    if svf_file then
        tmpfile:write("svf "..svf_file.."\n")
    end
    tmpfile:close()

    os.execute('jtag /tmp/jtag.cmd 2>&1 > /tmp/jtag.log')

    -- Restore GPIOs
    map:w32(0x00114, save_SICRL)
    map:w32(0x00118, save_SICRH)

    local f = assert(io.open('/tmp/jtag.log', 'r'))
    local results = f:read('*all')
    f:close()

    if results:find('error') then return results end
end

function fpga.configure_from_file(svf_file)
    return reprogram_fpga(svf_file)
end

function fpga.configure_from_spi_flash(svf_file)
    local tmp_svf_file = '/tmp/reload_fpga.svf'
    local tmpfile = io.open(tmp_svf_file, 'w+')
    local previous_line = ""
    for line in io.lines(svf_file) do
        tmpfile:write(line)
        tmpfile:write('\n')
        if previous_line:find('RUNTEST') then
            tmpfile:close()
            return reprogram_fpga(tmp_svf_file)
        else
            previous_line = line
        end
    end
    return "Invalid SVF file"
end

function fpga.verify_jtag()
    return reprogram_fpga()
end

-- Run JTAG commands, with UrJTAG
function fpga.jtag(commands)
    if not commands then
        return "No UrJTAG commands given"
    end
    local f = assert(io.popen("/sbin/lsmod | grep connctrl", 'r'))
    local connctrl = f:read()
    f:close()
    if connctrl then
        return "Cannot proceed, connctrl driver is loaded"
    end

    local fd = assert(u.open('/dev/mem', u.O_RDWR))
    local map = assert(u.mmap(0x100000, fd, 0xe0000000))
    assert(u.close(fd))

    local save_SICRL = map:r32(0x00114)
    local save_SICRH = map:r32(0x00118)

    -- Set USB_B group to GPIO (for GPIO2[11] => TMS)
    -- Set GPIO1_L to GPIO (for GPIO1[11] => TDO)
    -- Set GPIO1_J to GPIO (for GPIO1[10] => TDI)
    local new_SICRL = b.bor(b.band(save_SICRL, 0xcfff4fff), 0x00000000)

    -- Set GPIO2_F to GPIO (for GPIO2[17] => TCK)
    local new_SICRH = b.bor(b.band(save_SICRH, 0xfffffcff), 0x00000100)

    -- Reconfigure GPIOs
    map:w32(0x00114, new_SICRL)
    map:w32(0x00118, new_SICRH)

    local cmd_file = io.open('/tmp/urjtag.cmd', 'w+')
    cmd_file:write("cable gpio tdi=203 tdo=234 tck=209 tms=235\n")
    cmd_file:write("detect\n")
    cmd_file:write(commands)
    cmd_file:close()

    f = io.popen('jtag /tmp/urjtag.cmd', 'r')
    local results = f:read('*all')
    f:close()

    -- Restore GPIOs
    map:w32(0x00114, save_SICRL)
    map:w32(0x00118, save_SICRH)

    return results
end

return fpga

