--------------------------------------------------------------------------------
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--------------------------------------------------------------------------------
-- This module implements a linear (non-circular) expandable FIFO buffer. It
-- exposes methods to append new data and to peek at or consume existing data.
-- The buffer dynamically grows as needed to accomodate its contents. The
-- buffer capacity can optionally be capped at a specified maximum.

require 'middleclass'

require('lj_include') {
    'stdlib',
    'string'
}

local ffi = require 'ffi'
local C = ffi.C

local fifo = class('fifo')

function fifo:initialize(initial_capacity, max_capacity)
    if initial_capacity and max_capacity and initial_capacity > max_capacity then
        error('FIFO: Specified initial capacity is greater than the maximum')
    end
    self._size = 0
    self._capacity = initial_capacity or math.min(16 * 1024, max_capacity)
    self._max_capacity = max_capacity
    self._buf = ffi.cast('unsigned char *', C.malloc(self._capacity))
    if not self._buf then error('FIFO: Failed to allocate memory') end
end

function fifo:add(ptr, len)
    local needed = self._size + len
    if needed > self._capacity then
        while needed > self._capacity do
            self._capacity = self._capacity * 2
        end
        if self._max_capacity then
            self._capacity = math.min(self._capacity, self._max_capacity)
        end
        local new_buf = ffi.cast('unsigned char *', C.realloc(self._buf, self._capacity))
        if not new_buf then
            self:destroy()
            error('FIFO: Failed to expand buffer to new size of ' .. self._capacity)
        end
        self._buf = new_buf
    end
    len = math.min(len, self._capacity - self._size)
    if len > 0 then
        C.memcpy(self._buf + self._size, ptr, len)
        self._size = self._size + len
    end
    return len
end

function fifo:alloc_add()
    if self._size == self._capacity then
        self._capacity = self._capacity * 2
        if self._max_capacity then
            if self._size == self._max_capacity then return nil, 0 end
            self._capacity = math.min(self._capacity, self._max_capacity)
        end
        local new_buf = ffi.cast('unsigned char *', C.realloc(self._buf, self._capacity))
        if not new_buf then
            self:destroy()
            error('FIFO: Failed to expand buffer to new size of ' .. self._capacity)
        end
        self._buf = new_buf
    end
    return self._buf + self._size, self._capacity - self._size
end

function fifo:commit_add(len)
    self._size = self._size + len
end

function fifo:consume(len)
    len = math.min(len, self._size)
    if len > 0 then
        self._size = self._size - len
        if self._size > 0 then C.memmove(self._buf, self._buf + len, self._size) end
    end
    return len
end

function fifo:consume_line()
    local i = 0
    local buf = self:data()
    local size = self:size()
    while i < size do
        local ch = buf[i]
        if ch == 0x0A then
            local str = ffi.string(buf, i)
            self:consume(i + 1)
            return str
        elseif ch == 0x0D then
            local str = ffi.string(buf, i)
            if i + 1 < size and buf[i + 1] == 0x0A then
                self:consume(i + 2)
            else
                self:consume(i + 1)
            end
            return str
        end
        i = i + 1
    end
end

function fifo:data()
    return self._buf
end

function fifo:size()
    return self._size
end

function fifo:capacity()
    return self._capacity
end

function fifo:max_capacity()
    return self._max_capacity
end

function fifo:destroy()
    C.free(self._buf)
end

return fifo
