--------------------------------------------------------------------------------
-- Copyright (C) 2016 HBM, Inc. All rights reserved.
--------------------------------------------------------------------------------

require 'middleclass'
require 'pack'

local bit   = require 'bit'
local fifo  = require 'net/fifo'
local ffi   = require 'ffi'

--------------------------------------------------------------------------------

local ws_server = class('ws_server')

--------------------------------------------------------------------------------

local guid = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"

local sha1 = function(msg)
    local h0 = 0x67452301
    local h1 = 0xEFCDAB89
    local h2 = 0x98BADCFE
    local h3 = 0x10325476
    local h4 = 0xC3D2E1F0
    local bits = #msg * 8
    msg = msg .. string.char(0x80)
    local bytes = #msg + 8
    local fill_bytes = 64 - (bytes % 64)
    if fill_bytes ~= 64 then
        msg = msg .. string.rep(string.char(0), fill_bytes)
    end
    local high = math.floor(bits / 2^32)
    local low = bits - high * 2^32
    msg = msg .. string.pack('>I>I', high, low)
    for j = 1, #msg, 64 do
        local chunk = msg:sub(j, j + 63)
        local words = { string.unpack(chunk, string.rep('>I', 16)) }
        table.remove(words, 1)
        for i = 17, 80 do
            words[i] = bit.bxor(words[i - 3], words[i - 8], words[i - 14], words[i - 16])
            words[i] = bit.rol(words[i], 1)
        end
        local a = h0
        local b = h1
        local c = h2
        local d = h3
        local e = h4
        for i = 1, 80 do
            local k, f
            if i > 0 and i < 21 then
                f = bit.bor(bit.band(b, c), bit.band(bit.bnot(b), d))
                k = 0x5A827999
            elseif i > 20 and i < 41 then
                f = bit.bxor(b, c, d)
                k = 0x6ED9EBA1
            elseif i > 40 and i < 61 then
                f = bit.bor(bit.band(b, c), bit.band(b, d), bit.band(c, d))
                k = 0x8F1BBCDC
            elseif i > 60 and i < 81 then
                f = bit.bxor(b, c, d)
                k = 0xCA62C1D6
            end
            local temp = bit.rol(a, 5) + f + e + k + words[i]
            e = d
            d = c
            c = bit.rol(b,30)
            b = a
            a = temp
        end
        h0 = h0 + a
        h1 = h1 + b
        h2 = h2 + c
        h3 = h3 + d
        h4 = h4 + e
    end
    h0 = bit.band(h0, 0xffffffff)
    h1 = bit.band(h1, 0xffffffff)
    h2 = bit.band(h2, 0xffffffff)
    h3 = bit.band(h3, 0xffffffff)
    h4 = bit.band(h4, 0xffffffff)
    return string.pack('>i>i>i>i>i', h0, h1, h2, h3, h4)
end

local base64chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/'

local base64_encode = function(data)
    local result = ''
    local padding = ''
    local count = #data % 3
    if count > 0 then
        for i = count, 2 do
            padding = padding .. '='
            data = data .. '\0'
        end
    end
    local bytes = 0
    for i = 1, #data, 3 do
        local chars = { data:sub(i, i + 2):byte(1, 3) }
        local n = bit.lshift(chars[1], 16) + bit.lshift(chars[2], 8) + chars[3]
        local narr = {}
        narr[1] = bit.band(bit.rshift(n, 18), 63) + 1
        narr[2] = bit.band(bit.rshift(n, 12), 63) + 1
        narr[3] = bit.band(bit.rshift(n, 6), 63) + 1
        narr[4] = bit.band(n, 63) + 1
        result = result .. base64chars:sub(narr[1], narr[1])
        result = result .. base64chars:sub(narr[2], narr[2])
        result = result .. base64chars:sub(narr[3], narr[3])
        result = result .. base64chars:sub(narr[4], narr[4])
    end
    return result:sub(1, #result - #padding) .. padding
end

local function get_ws_accept_key(key)
    local a = key .. guid
    local sha1 = sha1(a)
    return base64_encode(sha1)
end

--------------------------------------------------------------------------------

function ws_server:initialize(stream)

    self.stream = stream
    self.rxq = fifo:new(16 * 1024)
    self.established = false
    self.headers = {}

    stream:on_data(function(...) self:_on_data(...) end)
    stream:on_close(function(...) self:_on_close(...) end)
    stream:on_error(function(...) self:_on_error(...) end)

end

--------------------------------------------------------------------------------

function ws_server:_on_data()

    if not self.established then

        while true do

            local line = self.stream:fifo():consume_line()
            if not line then break end

            if not self.path then

                local method, path, version = line:match('^(%S+)%s+(%S+)%s+(%S+)$')
                if not method or not path or not version then
                    self.stream:close()
                    return
                end
                self.path = path
                self.method = method

            elseif line == '' then

                if not self.path or
                        not self.headers['Sec-WebSocket-Protocol'] or
                        not self.headers['Sec-WebSocket-Key'] or
                        not self.headers['Connection'] then
                    self.stream:close()
                    return
                end

                self.stream:send("HTTP/1.1 101 Switching Protocols\r\n")
                self.stream:send("Upgrade: websocket\r\n")
                self.stream:send(string.format("Connection: %s\r\n",
                    self.headers['Connection']))
                self.stream:send(string.format("Sec-Websocket-Accept: %s\r\n",
                    get_ws_accept_key(self.headers['Sec-WebSocket-Key'])))
                self.stream:send(string.format("Sec-Websocket-Protocol: %s\r\n",
                    self.headers['Sec-WebSocket-Protocol']))
                self.stream:send("\r\n")

                self.established = true
                if self.on_handshake then self.on_handshake(self) end

            else

                local key, value = line:match('^(%S+):%s+(.+)$')
                self.headers[key] = value

            end

        end

    else

        local fifo = self.stream:fifo()
        local buf = fifo:data()
        local n = fifo:size()
        local i = 0
        while i + 2 <= n do
            local flags = bit.rshift(buf[i], 4)
            local opcode = bit.band(buf[i], 15)
            local mask = bit.rshift(buf[i + 1], 7)
            local payload_size = bit.band(buf[i + 1], 127)
            local header_size = 2
            if payload_size == 126 then
                if i + 4 > n then break end
                payload_size = bit.lshift(buf[i + 2], 8) + buf[i + 3]
                header_size = header_size + 2
            elseif payload_size == 127 then
                if i + 10 > n then break end
                payload_size = bit.lshift(buf[i + 6], 24) +
                    bit.lshift(buf[i + 7], 16) +
                    bit.lshift(buf[i + 8], 8) +
                    buf[i + 9]
                header_size = header_size + 8
            end
            if i + header_size + payload_size > n then break end
            self.rxq:add(buf + i + header_size, payload_size)
            if self.on_data then self.on_data(self) end
            i = i + header_size + payload_size
        end
        if i > 0 then fifo:consume(i) end

    end

end

--------------------------------------------------------------------------------

function ws_server:_on_close()
    if self.rxq then
        self.rxq:destroy()
        self.rxq = nil
    end
    if self.on_close then self.on_close(self) end
end

--------------------------------------------------------------------------------

function ws_server:_on_error(message)
    if self.on_error then self.on_error(self, message) end
end

--------------------------------------------------------------------------------

function ws_server:on_data(handler) self.on_data = handler end
function ws_server:on_close(handler) self.on_close = handler end
function ws_server:on_error(handler) self.on_error = handler end
function ws_server:on_handshake(handler) self.on_handshake = handler end

--------------------------------------------------------------------------------

function ws_server:fifo()
    return self.rxq
end

--------------------------------------------------------------------------------

function ws_server:close()
    self.stream:close()
end

--------------------------------------------------------------------------------

function ws_server:send(ptr, len)
    local header = ffi.new('unsigned char[10]')
    header[0] = 0x82 -- binary frame
    if len < 126 then -- 7-bit length format
        header[1] = len
        self.stream:send(header, 2)
    elseif len <= 65535 then -- 16-bit length format
        header[1] = 126
        header[2] = bit.rshift(len, 8)
        header[3] = bit.band(len, 255)
        self.stream:send(header, 4)
    else
        header[1] = 127
        header[2] = 0
        header[3] = 0
        header[4] = 0
        header[5] = 0
        header[6] = bit.rshift(len, 24)
        header[7] = bit.rshift(len, 16)
        header[8] = bit.rshift(len, 8)
        header[9] = bit.band(len, 255)
        self.stream:send(header, 10)
    end
    self.stream:send(ptr, len)
end

--------------------------------------------------------------------------------

return ws_server
